@@ -40,7 +40,7 @@
 
 #include <reentrant.h>
 #include <sys/types.h>
-#include <sys/poll.h>
+#include <poll.h>
 #include <assert.h>
 #include <errno.h>
 #include <stdlib.h>
@@ -55,9 +55,6 @@
 
 #define	RQCRED_SIZE	400	/* this size is excessive */
 
-#define SVC_VERSQUIET 0x0001	/* keep quiet about vers mismatch */
-#define version_keepquiet(xp) ((u_long)(xp)->xp_p3 & SVC_VERSQUIET)
-
 #define max(a, b) (a > b ? a : b)
 
 /*
@@ -100,16 +97,43 @@ xprt_register (xprt)
   rwlock_wrlock (&svc_fd_lock);
   if (__svc_xports == NULL)
     {
-      __svc_xports = (SVCXPRT **) mem_alloc (FD_SETSIZE * sizeof (SVCXPRT *));
+      __svc_xports = (SVCXPRT **) calloc (_rpc_dtablesize(), sizeof (SVCXPRT *));
       if (__svc_xports == NULL)
 	return;
-      memset (__svc_xports, '\0', FD_SETSIZE * sizeof (SVCXPRT *));
     }
-  if (sock < FD_SETSIZE)
+  if (sock < _rpc_dtablesize())
     {
+      int i;
+      struct pollfd *new_svc_pollfd;
+
       __svc_xports[sock] = xprt;
-      FD_SET (sock, &svc_fdset);
-      svc_maxfd = max (svc_maxfd, sock);
+      if (sock < FD_SETSIZE)
+	{
+          FD_SET (sock, &svc_fdset);
+	  svc_maxfd = max (svc_maxfd, sock);
+	}
+
+      /* Check if we have an empty slot */
+      for (i = 0; i < svc_max_pollfd; ++i)
+        if (svc_pollfd[i].fd == -1)
+          {
+            svc_pollfd[i].fd = sock;
+            svc_pollfd[i].events = (POLLIN | POLLPRI |
+                                    POLLRDNORM | POLLRDBAND);
+            return;
+          }
+
+      new_svc_pollfd = (struct pollfd *) realloc (svc_pollfd,
+                                                  sizeof (struct pollfd)
+                                                  * (svc_max_pollfd + 1));
+      if (new_svc_pollfd == NULL) /* Out of memory */
+        return;
+      svc_pollfd = new_svc_pollfd;
+      ++svc_max_pollfd;
+
+      svc_pollfd[svc_max_pollfd - 1].fd = sock;
+      svc_pollfd[svc_max_pollfd - 1].events = (POLLIN | POLLPRI |
+                                               POLLRDNORM | POLLRDBAND);
     }
   rwlock_unlock (&svc_fd_lock);
 }
@@ -142,16 +166,25 @@ __xprt_do_unregister (xprt, dolock)
 
   if (dolock)
     rwlock_wrlock (&svc_fd_lock);
-  if ((sock < FD_SETSIZE) && (__svc_xports[sock] == xprt))
+  if ((sock < _rpc_dtablesize() ) && (__svc_xports[sock] == xprt))
     {
+      int i;
+
       __svc_xports[sock] = NULL;
-      FD_CLR (sock, &svc_fdset);
-      if (sock >= svc_maxfd)
+      if (sock < FD_SETSIZE)
 	{
-	  for (svc_maxfd--; svc_maxfd >= 0; svc_maxfd--)
-	    if (__svc_xports[svc_maxfd])
-	      break;
+          FD_CLR (sock, &svc_fdset);
+	  if (sock >= svc_maxfd)
+       	    {
+              for (svc_maxfd--; svc_maxfd >= 0; svc_maxfd--)
+                if (__svc_xports[svc_maxfd])
+                  break;
+            }
 	}
+
+      for (i = 0; i < svc_max_pollfd; ++i)
+        if (svc_pollfd[i].fd == sock)
+          svc_pollfd[i].fd = -1;
     }
   if (dolock)
     rwlock_unlock (&svc_fd_lock);
@@ -467,20 +500,14 @@ void
 __svc_versquiet_on (xprt)
      SVCXPRT *xprt;
 {
-  u_long tmp;
-
-  tmp = ((u_long) xprt->xp_p3) | SVC_VERSQUIET;
-  xprt->xp_p3 = tmp;
+  svc_flags (xprt) |= SVC_VERSQUIET;
 }
 
 void
 __svc_versquiet_off (xprt)
      SVCXPRT *xprt;
 {
-  u_long tmp;
-
-  tmp = ((u_long) xprt->xp_p3) & ~SVC_VERSQUIET;
-  xprt->xp_p3 = tmp;
+  svc_flags (xprt) &= ~SVC_VERSQUIET;
 }
 
 void
@@ -494,7 +521,7 @@ int
 __svc_versquiet_get (xprt)
      SVCXPRT *xprt;
 {
-  return ((int) xprt->xp_p3) & SVC_VERSQUIET;
+  return version_keepquiet (xprt);
 }
 #endif
 
@@ -606,11 +633,15 @@ svc_getreqset (readfds)
   int bit, fd;
   fd_mask mask, *maskp;
   int sock;
+  int setsize;
 
   assert (readfds != NULL);
 
+  setsize = _rpc_dtablesize ();
+  if (setsize > FD_SETSIZE)
+    setsize = FD_SETSIZE;
   maskp = readfds->fds_bits;
-  for (sock = 0; sock < FD_SETSIZE; sock += NFDBITS)
+  for (sock = 0; sock < setsize; sock += NFDBITS)
     {
       for (mask = *maskp++; (bit = ffsl(mask)) != 0; mask ^= (1L << (bit - 1)))
 	{
@@ -649,6 +680,7 @@ svc_getreq_common (fd)
     {
       if (SVC_RECV (xprt, &msg))
 	{
+	  bool_t no_dispatch;
 
 	  /* now find the exported program and call it */
 	  struct svc_callout *s;
@@ -660,11 +692,14 @@ svc_getreq_common (fd)
 	  r.rq_proc = msg.rm_call.cb_proc;
 	  r.rq_cred = msg.rm_call.cb_cred;
 	  /* first authenticate the message */
-	  if ((why = _authenticate (&r, &msg)) != AUTH_OK)
+	  why = _gss_authenticate(&r, &msg, &no_dispatch);
+	  if (why != AUTH_OK)
 	    {
 	      svcerr_auth (xprt, why);
 	      goto call_done;
 	    }
+	  if (no_dispatch)
+	    goto call_done;
 	  /* now match message with a registered service */
 	  prog_found = FALSE;
 	  low_vers = (rpcvers_t) - 1L;
@@ -714,11 +749,6 @@ svc_getreq_common (fd)
 	  SVC_DESTROY (xprt);
 	  break;
 	}
-    else if ((xprt->xp_auth != NULL) &&
-	     (xprt->xp_auth->svc_ah_private == NULL))
-	{
-	  xprt->xp_auth = NULL;
-	}
     }
   while (stat == XPRT_MOREREQS);
 }
@@ -729,36 +759,22 @@ svc_getreq_poll (pfdp, pollretval)
      struct pollfd *pfdp;
      int pollretval;
 {
-  int i;
-  int fds_found;
+  int fds_found, i;
 
-  for (i = fds_found = 0; fds_found < pollretval; i++)
+  for (i = fds_found = 0; i < svc_max_pollfd; ++i)
     {
       struct pollfd *p = &pfdp[i];
 
-      if (p->revents)
+      if (p->fd != -1 && p->revents)
 	{
-	  /* fd has input waiting */
-	  fds_found++;
-	  /*
-	   *      We assume that this function is only called
-	   *      via someone _select()ing from svc_fdset or
-	   *      _poll()ing from svc_pollset[].  Thus it's safe
-	   *      to handle the POLLNVAL event by simply turning
-	   *      the corresponding bit off in svc_fdset.  The
-	   *      svc_pollset[] array is derived from svc_fdset
-	   *      and so will also be updated eventually.
-	   *
-	   *      XXX Should we do an xprt_unregister() instead?
-	   */
-	  if (p->revents & POLLNVAL)
-	    {
-	      rwlock_wrlock (&svc_fd_lock);
-	      FD_CLR (p->fd, &svc_fdset);
-	      rwlock_unlock (&svc_fd_lock);
-	    }
-	  else
-	    svc_getreq_common (p->fd);
+          /* fd has input waiting */
+          if (p->revents & POLLNVAL)
+	    xprt_unregister (__svc_xports[p->fd]);
+          else
+            svc_getreq_common (p->fd);
+
+          if (++fds_found >= pollretval)
+            break;
 	}
     }
 }
